﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Collections;

namespace TTT
{
    class HapticResponse
    {
        /// <summary>
        /// Calculates the appropriate haptic response for different materials
        /// </summary>

        #region Constants
        
        const double HARD_LOW = 0.2;
        const double HARD_HIGH = 3.5;

        const double SOFT_LOW = 0.3;
        const double SOFT_HIGH = 0.8;
        
        const double BREAKY_STEPS = 5;
        const double BREAKY_STEPS_FINE = 10;

        const double VERYSOFT_LOW = 0.1;
        const double VERYSOFT_HIGH = 0.2;

        const double BUTTON_PRESS_HARD_A = 0.75;
        const double BUTTON_PRESS_HARD_B = 0.98;

        const double BUTTON_PRESS_SOFT_A = 0.75;
        const double BUTTON_PRESS_SOFT_B = 0.90;


        /// <summary>
        /// The buffer is used to smoothen the signals
        /// </summary>
        private Queue<double> pressureBuffer = new Queue<double>();
        private int pressureBufferLength = 15;
        
        #endregion



        #region Color codes

        public static readonly Color COLORCODE_HARD         = Color.FromArgb(255, 0, 0);    // red
        public static readonly Color COLORCODE_NONE         = Color.FromArgb(0, 0, 0);      // black
        public static readonly Color COLORCODE_SOFT         = Color.FromArgb(0, 255, 0);    // green
        public static readonly Color COLORCODE_VERYSOFT     = Color.FromArgb(255, 255, 0);  // yellow
        public static readonly Color COLORCODE_BREAKY       = Color.FromArgb(0, 0, 255);    // blue

        #endregion



        #region Variables
        HapticResponses pastResponse = HapticResponses.None;

        //memory

        #endregion


        /// <summary>
        /// Constructor
        /// </summary>
        public HapticResponse()
        {
        }


        /// <summary>
        /// Calculate the linear response feedback
        /// </summary>
        /// <param name="low"></param>
        /// <param name="high"></param>
        /// <param name="pressure"></param>
        /// <returns></returns>
        private double LinearResponse(double low, double high, double pressure)
        {
            if (pressure < low)
            {
                return 1;
            }
            else if (pressure > high)
            {
                return 0;
            }
            else
            {
                return ((-pressure / high) + 1.0) / (1.0 - (low / high));
            }
        }



        public void ResetBuffer()
        {
            this.pressureBuffer.Clear();
        }



        /// <summary>
        /// Get a haptic response 
        /// </summary>
        /// <param name="response"></param>
        /// <param name="pressure"></param>
        /// <param name="position"></param>
        /// <returns></returns>
        public double GetHapticResponse(HapticResponses response, double pressureRAW, double position)
        {
            this.pressureBuffer.Enqueue(pressureRAW);
            if (this.pressureBuffer.Count > this.pressureBufferLength)
            {
                this.pressureBuffer.Dequeue();
            }

            double pressure = 0;
            foreach (double value in this.pressureBuffer)
            {
                pressure += value;
            }
            pressure = pressure / this.pressureBuffer.Count;

            double alpha, beta;

            switch (response)
            {
                case HapticResponses.Hard:
                    return LinearResponse(HARD_LOW, HARD_HIGH, pressure);
                    break;
                case HapticResponses.Soft:
                    return LinearResponse(SOFT_LOW, SOFT_HIGH, pressure);
                    break;
                case HapticResponses.VerySoft:
                    //return LinearResponse(VERYSOFT_LOW, VERYSOFT_HIGH, pressure);
                    double upscaled = BREAKY_STEPS_FINE * pressure;
                    double truncated = Math.Truncate(upscaled);
                    double saw = upscaled - truncated;
                    return 1.0 - (truncated / BREAKY_STEPS_FINE) + saw / BREAKY_STEPS_FINE;
                    break;
                case HapticResponses.Breaky:
                    double upscaledFine = BREAKY_STEPS * pressure;
                    double truncatedFine = Math.Truncate(upscaledFine);
                    double sawFine = upscaledFine - truncatedFine;
                    return 1.0 - (truncatedFine / BREAKY_STEPS) + sawFine / BREAKY_STEPS;
                    break;
                case HapticResponses.None:
                    return 1.0;
                    break;
                case HapticResponses.ButtonPressHard:
                    alpha = -4.0 * (BUTTON_PRESS_HARD_A - 1.0) / (4.0 * Math.Pow(BUTTON_PRESS_HARD_B, 2));
                    beta = -alpha * 2.0 * BUTTON_PRESS_HARD_B;
                    if (pressure > BUTTON_PRESS_HARD_B) return 0.0;
                    else return alpha * Math.Pow(pressure, 2) + beta * pressure + 1.0;
                    break;
                case HapticResponses.ButtonPressSoft:
                    alpha = -4.0 * (BUTTON_PRESS_SOFT_A - 1.0) / (4.0 * Math.Pow(BUTTON_PRESS_SOFT_B, 2));
                    beta = -alpha * 2.0 * BUTTON_PRESS_SOFT_B;
                    if (pressure > BUTTON_PRESS_SOFT_B) return 0.0;
                    else return alpha * Math.Pow(pressure, 2) + beta * pressure + 1.0;
                    break;
                case HapticResponses.ButtonPressVerySoft:
                    if (pressure > 0.7) return 0.0;
                    else return LinearResponse(SOFT_LOW, SOFT_HIGH, pressure);
                    break;
                case HapticResponses.Paint:
                    if (pressure > 0.7) return 0.5;
                    else return 1.0;
                    break;
                case HapticResponses.ButtonPressLayers:
                    double upscaledLayers = 3 * pressure;
                    double truncatedLayers = Math.Truncate(upscaledLayers);
                    double sawLayers = upscaledLayers - truncatedLayers;
                    return 1.0 - (truncatedLayers / 3) + sawLayers / 3;
                    break;
                case HapticResponses.TwoLayers:
                    if (pressure > 0.98) return 0.4;
                    else if (pressure > 0.75) return 0.7;
                    else return 1.0;
                    break;
                default:
                    break;
            }


            return pressure;
        }


    }


    /// <summary>
    /// The currently implemented haptic responses
    /// </summary>
    public enum HapticResponses
    {
        None,
        Hard,
        Soft,
        VerySoft,
        Breaky,
        ButtonPressHard,
        ButtonPressSoft,
        ButtonPressVerySoft,
        ButtonPressLayers,
        Sluggish,
        Paint,
        TwoLayers
    }
}
